"""
Module with functions to transform programs.

Classes:
ProgramTransformer -- Class to transform programs.
"""

from . import transformer as _tf
from . import term as _tt
from . import head as _th

import clingo as _clingo
from clingo import ast as _ast

class ProgramTransformer(_ast.Transformer):
    """
    Rewrites all temporal operators in a logic program.

    Statements should be passed one after the other to the visit method.

    Members:
    __final            -- Final rules are put into the always program part and
                          the __final atom put into their body. This flag
                          indicates that the __final atom has to be appended.
    __head             -- Indicates that the head of a rule is being visited.
    __constraint       -- Whether the current statement is a constraint.
    __negation         -- Whether the child nodes are in the scope of a
                          negation (this is overridden by positive nested
                          literals).
    __normal           -- Whether the current statement is a normal rule.
    __max_shift        -- The maximum number of steps a rule looks into the
                          future. Determines window to reground constraints.
                          Stored as a list with one integer element to allow
                          passing by reference.
    __term_transformer -- The transformer used to rewrite terms.
    __constraint_parts -- Parts that have to be regrounded because of
                          constraints referring to the future.
    __aux_rules        -- Auxiliary always quantified rules added during
                          translation.
    """
    def __init__(self, future_predicates, constraint_parts, aux_rules):
        self.__final = False
        self.__head = False
        self.__constraint = False
        self.__negation = False
        self.__normal = False
        self.__max_shift = [0]
        self.__term_transformer = _tt.TermTransformer(future_predicates)
        self.__head_transformer = _th.HeadTransformer()
        self.__constraint_parts = constraint_parts
        self.__aux_rules        = aux_rules

    def __append_final(self, x, param=None):
        loc = x.location
        fun = _ast.Function(loc, "__final", [_ast.SymbolicTerm(loc, param)] if param is not None else [], False)
        x.body.append(_ast.Literal(loc, _ast.Sign.NoSign, _ast.SymbolicAtom(fun)));

    def visit(self, x, *args, **kwargs):
        """
        Extends the transformer's generic visit method to add the final atom to
        all AST nodes in final program parts having a body.

        The extension happens before the node is visited normally so the time
        parameter is added to the atom accordingly.
        """
        if self.__final and isinstance(x, _ast.AST) and hasattr(x, "body"):
            self.__append_final(x)
        if isinstance(x, _ast.ASTSequence):
            return _ast.Transformer.visit_sequence(self, x, *args, **kwargs)

        return _ast.Transformer.visit(self, x, *args, **kwargs)

    def visit_Rule(self, rule):
        """
        Sets the state flags when visiting a rule.

        After that the head and body of the rule are visited in the right context.
        """
        try:
            self.__head = True
            self.__max_shift = [0]
            self.__constraint = _tf.is_constraint(rule)
            self.__normal = _tf.is_normal(rule)
            rule.head = self.visit(rule.head)
            self.__head = False
            rule.body = self.visit(rule.body)
            if self.__max_shift[0] > 0 and not self.__final:
                last = _ast.Rule(rule.location, rule.head, rule.body[:])
                self.__append_final(rule, _clingo.Function(_tf.g_time_parameter_name_alt))
                self.__constraint_parts.setdefault((self.__part, self.__max_shift[0]), []).append((rule, last))
                return None
        finally:
            self.__head        = False
            self.__max_shift   = [0]
            self.__constraint  = False
            self.__normal = False
        return rule

    def visit_Literal(self, literal):
        """
        Removes the head flag for negative head literals.
        """
        head = self.__head
        try:
            self.__negation = literal.sign != _ast.Sign.NoSign
            self.__head = self.__head and literal.sign == _ast.Sign.NoSign
            return literal.update(**self.visit_children(literal))
        finally:
            self.__negation = False
            self.__head = head

    def visit_ConditionalLiteral(self, literal):
        """
        Make sure that conditions are traversed as non-head literals.
        """
        self.visit(literal.literal)
        head = self.__head
        try:
            self.__head = False
            self.visit(literal.condition)
        finally:
            self.__head = head
        return literal

    def visit_SymbolicAtom(self, atom):
        """
        Rewrites the given symbolic atom appending a time parameter.

        If this atom appears in a head then it is also replaced by a
        corresponding future atom defined later.
        """
        atom.symbol = self.__term_transformer.visit(atom.symbol, self.__head, not self.__constraint and (not self.__head or not self.__normal), self.__head, self.__max_shift)
        return atom

    def visit_TheoryAtom(self, atom):
        """
        Rewrites theory atoms related to temporal formulas.

        An atom of form `&tel {...}` is rewritten to `&tel(k) {...}`, atoms of
        form `&initial` and `&final` are rewritten to `__initial` and
        `__final`, and atoms of form `&true` and `&false` are rewritten to
        `#true` and `#false`.
        """
        if atom.term.ast_type == _ast.ASTType.Function and len(atom.term.arguments) == 0:
            time = lambda loc: _ast.SymbolicTerm(loc, _clingo.Function(_tf.g_time_parameter_name))
            wrap = lambda loc, atom: _ast.Literal(loc, _ast.Sign.DoubleNegation, atom) if self.__head else atom
            if atom.term.name == "del" :
                if not self.__negation and not self.__constraint:
                        raise RuntimeError("dynamic formulas not supported in this context: {}".format(_tf.str_location(atom.location)))
                atom.term.arguments = [_ast.SymbolicTerm(atom.term.location, _clingo.Function("__t"))]
            elif atom.term.name == "tel" :
                if self.__head:
                    atom, rules = self.__head_transformer.transform(atom)
                    self.__aux_rules.extend(rules)
                else:
                    if not self.__negation and not self.__constraint:
                        raise RuntimeError("temporal formulas not supported in this context: {}".format(_tf.str_location(atom.location)))
                    for element in atom.elements:
                        if len(element.terms) != 1:
                            raise RuntimeError("invalid temporal formula: {}".format(_tf.str_location(atom.location)))
                        self.visit(element.condition)
                    atom.term = self.__term_transformer.visit(atom.term, False, True, True, self.__max_shift)
            elif atom.term.name == "initial":
                atom = wrap(atom.location, _ast.SymbolicAtom(_ast.Function(atom.location, "__initial", [time(atom.location)], False)))
            elif atom.term.name == "final":
                atom = wrap(atom.location, _ast.SymbolicAtom(_ast.Function(atom.location, "__final", [time(atom.location)], False)))
            elif atom.term.name == "true":
                atom = wrap(atom.location, _ast.BooleanConstant(True))
            elif atom.term.name == "false":
                atom = wrap(atom.location, _ast.BooleanConstant(False))
        return atom

    def visit_Program(self, prg):
        """
        Adds the time parameter to the given program given directive.

        Furthermore, the final program part is turned into an always program
        part and the __final flag set accordingly.
        """
        self.__final = prg.name == "final"
        if self.__final:
            prg.name = "always"
        if prg.name == "base":
            prg.name = "initial"
        prg.parameters.append(_ast.Id(prg.location, _tf.g_time_parameter_name))
        prg.parameters.append(_ast.Id(prg.location, _tf.g_time_parameter_name_alt))
        self.__part = prg.name
        return prg

    def visit_ShowSignature(self, sig):
        """
        Adjusts the arity of show predicate statements.

        For example `#show p/2` becomes `#show p/3` because all occurrences of
        atoms over `p` are extended with a time parameter.
        """
        sig.arity += 1
        return sig

    def visit_ProjectSignature(self, sig):
        """
        Adjusts the arity of project predicate statements.

        See visit_ShowSignature.
        """
        sig.arity += 1
        return sig

    def visit_Input(self, sig):
        """
        Adjusts the arity of input predicate statements.

        See visit_ShowSignature.
        """
        sig.arity += 1
        return sig






"""
The telingo module contains functions to translate and solve temporal logic
programs.

Classes:
Application -- Main application class.

Functions:
imain -- Function to run the incremetal solving loop.
main  -- Main function starting an extended clingo application.
"""

from . import transformers as _tf
from . import theory as _ty

import sys as _sys
import clingo as _clingo
import clingo.ast as _ast
import textwrap as _textwrap

def imain(prg, future_sigs, program_parts, on_model, imin = 0, imax = None, istop = "SAT"):
    """
    Take a program object and runs the incremental main solving loop.

    For each pair (name, arity) in future_sigs all atoms in the program base
    with the time parameter referring to the future are set to false. For
    example, given (p, 2) and atoms  p(x,1) in step 0, the atom would p(x,1)
    would be set to false via an assumption. In the following time steps, it
    would not be set to False.

    The list program_parts contains all program parts appearing in the program
    in form of triples (root, name, range) where root is either "initial" (time
    step 0), "always" (time steps >= 0), or "dynamic" (time steps > 0) and
    range is a list of integers for which the part has to be grounded
    backwards. Given range [0, 1] and root "always", at each iteration the
    program part would be grounded at horizon and horizon-1. The latter only if
    the horizon is greater than 0.

    Arguments:
    prg           -- Control object holding the program.
    future_sigs   -- Signatures of predicates whose future incarnations have to
                     be set to False.
    program_parts -- Program parts to ground.
    imin          -- Minimum number of iterations.
    imax          -- Maximum number of iterations.
    istop         -- When to stop.
    """
    f = _ty.Theory()
    step, ret = 0, None
    while ((imax is None or step < imax) and
           (step == 0 or step < imin or (
              (istop == "SAT"     and not ret.satisfiable) or
              (istop == "UNSAT"   and not ret.unsatisfiable) or
              (istop == "UNKNOWN" and not ret.unknown)))):
        parts = []
        for root_name, part_name, rng in program_parts:
            for i in rng:
                if ((step - i >= 0 and root_name == "always") or
                    (step - i  > 0 and root_name == "dynamic") or
                    (step - i == 0 and root_name == "initial")):
                    parts.append((part_name, [_clingo.Number(step - i), _clingo.Number(step)]))
        if step > 0:
            prg.release_external(_clingo.Function("__final", [_clingo.Number(step-1)]))
            prg.cleanup()
        prg.ground(parts)
        f.translate(step, prg)
        prg.assign_external(_clingo.Function("__final", [_clingo.Number(step)]), True)
        assumptions = []
        for name, arity, positive in future_sigs:
            for atom in prg.symbolic_atoms.by_signature(name, arity, positive):
                if atom.symbol.arguments[-1].number > step:
                    assumptions.append(-atom.literal)
        ret, step = prg.solve(on_model=lambda m: on_model(m, step), assumptions=assumptions), step+1

class Application:
    """
    Application object as accepted by clingo.clingo_main().

    Rewrites the incoming temporal logic programs into incremental ASP programs
    and solves them.
    """
    def __init__(self):
        """
        Initializes the application setting the program name.

        See clingo.clingo_main().
        """
        self.program_name = "telingo"
        self.version = "2.1.1"

        self.__imin = 0
        self.__imax = None
        self.__istop = "SAT"
        self.__horizon = 0

    def __on_model(self, model, horizon):
        """
        Prints the atoms in a model grouped by state.

        Arguments:
        model -- The model to print.
        horizon -- The number of states.
        """
        self.__horizon = horizon

    def __parse_imin(self, value):
        """
        Parse imin argument.
        """
        self.__imin = int(value)
        return self.__imin >= 0

    def __parse_imax(self, value):
        """
        Parse imax argument.
        """
        if len(value) > 0:
            self.__imax = int(value)
            return self.__imax >= 0
        else:
            self.__imax = None
            return True

    def __parse_istop(self, value):
        """
        Parse istop argument.
        """
        self.__istop = value.upper()
        return self.__istop in ["SAT", "UNSAT", "UNKNOWN"]

    def print_model(self, model, printer):
        table = {}
        for sym in model.symbols(shown=True):
            if sym.type == _clingo.SymbolType.Function and len(sym.arguments) > 0:
                table.setdefault(sym.arguments[-1].number, []).append(_clingo.Function(sym.name, sym.arguments[:-1], sym.positive))
        for step in range(self.__horizon+1):
            symbols = table.get(step, [])
            _sys.stdout.write(" State {}:".format(step))
            sig = None
            for sym in sorted(symbols):
                if not sym.name.startswith('__'):
                    if (sym.name, len(sym.arguments), sym.positive) != sig:
                        _sys.stdout.write("\n ")
                        sig = (sym.name, len(sym.arguments), sym.positive)
                    _sys.stdout.write(" {}".format(sym))
            _sys.stdout.write("\n")
        return True

    def register_options(self, options):
        """
        See clingo.clingo_main().
        """
        group = "Telingo Options"
        options.add(group, "imin", "Minimum number of solving steps [0]", self.__parse_imin, argument="<n>")
        options.add(group, "imax", "Maximum number of solving steps []", self.__parse_imax, argument="<n>")
        options.add(group, "istop", _textwrap.dedent("""\
            Stop criterion [sat]
                  <arg>: {sat|unsat|unknown}"""), self.__parse_istop)

    def main(self, prg, files):
        """
        Implements the incremental solving loop.

        This function implements the Application.main() function as required by
        clingo.clingo_main().
        """
        with _ast.ProgramBuilder(prg) as b:
            files = [open(f) for f in files]
            if len(files) == 0:
                files.append(_sys.stdin)
            future_sigs, program_parts = _tf.transform([f.read() for f in files], b.add)

        imain(prg, future_sigs, program_parts, self.__on_model, self.__imin, self.__imax, self.__istop)

def main():
    """
    Run the telingo application.
    """
    _sys.exit(int(_clingo.clingo_main(Application(), _sys.argv[1:])))



"""
This module exports functions to translate (ground) theory atoms occuring in
rule heads to rules via clingo's backend.
"""

import clingo as _clingo
from telingo.transformers import transformer as _tf
from collections import namedtuple as _namedtuple
from . import body as _bd
from .formula import *
import itertools as _it
import functools as _ft
from clingo import ast as _ast


def new_tuple(name, fields, keys, tostring=None):
    ret = _namedtuple(name, fields)
    ret.child_keys = keys
    ret.ast_type = name
    ret.keys = fields
    if tostring is not None:
        ret.__str__ = tostring
    ret._rep = property(ret.__str__, "get string representation of tuple")
    return ret


class FormulaToStr(_tf.TelTransformer):
    """
    Converts head formuals to string.
    """
    def visit_TelAtom(self, x):
        args = "" if len(x.arguments) == 0 else "({})".format(",".join(map(str, x.arguments)))
        sign = "" if x.positive else "-"
        return "{}{}{}".format(sign, x.name, args)

    def visit_TelNext(self, x):
        op = ">:" if x.weak else ">"
        return "({}{}{})".format(x.lhs, op, self(x.rhs))

    def visit_TelUntil(self, x):
        op = ">?" if x.until else ">*"
        lhs = "" if x.lhs is None else self(x.lhs)
        return "({}{}{})".format(lhs, op, self(x.rhs))

    def visit_TelClause(self, x):
        if len(x.elements) == 1:
            return self(x.elements[0])
        op = "&" if x.conjunctive else "|"
        return "({})".format(op.join(map(self, x.elements)))

    def visit_TelNegation(self, x):
        return "(~{})".format(self(x.rhs))

    def visit_TelConstant(self, x):
        return "&true" if x.value else "&false"

    def visit_TelShift(self, x):
        if x.lhs == 0:
            return "(~(~{})".format(x.rhs)
        elif x.lhs < 0:
            return "(~(~({}<{}))".format(-x.lhs, x.rhs)
        elif x.lhs > 0:
            return "(~(~({}>{}))".format(x.lhs, x.rhs)

def formula_to_str(x):
    return FormulaToStr()(x)

TelNext = new_tuple("TelNext", ["lhs", "rhs", "weak"], ["rhs"], formula_to_str)
TelUntil = new_tuple("TelUntil", ["lhs", "rhs", "until"], ["lhs", "rhs"], formula_to_str)
TelAtom = new_tuple("TelAtom", ["positive", "name", "arguments"], ["arguments"], formula_to_str)
TelClause = new_tuple("TelClause", ["elements", "conjunctive"], ["elements"], formula_to_str)
TelNegation = new_tuple("TelNegation", ["rhs"], ["rhs"], formula_to_str)
TelConstant = new_tuple("TelConstant", ["value"], [], formula_to_str)
TelShift = new_tuple("TelShift", ["lhs", "rhs"], [], formula_to_str)


def create_atom(rep, add_formula, positive):
    """
    Returns the atom corresponding the given theory term.

    Throws an error if rep it is not a valid atom.

    Arguments:
    rep         -- Theory term to translate.
    add_formula -- Callback to add resulting formulas.
    positive    -- Boolean indicating the classical sign.
    """
    if rep.type == _clingo.TheoryTermType.Symbol:
        return add_formula(TelAtom(positive, rep.name, []))
    elif rep.type == _clingo.TheoryTermType.Function:
        if rep.name == "-" and len(rep.arguments) == 1:
            return create_atom(rep.arguments[0], add_formula, not positive)
        elif rep.name not in g_binary_operators and rep.name not in g_unary_operators and rep.name not in g_tel_operators and rep.name not in g_arithmetic_operators:
            return add_formula(TelAtom(positive, rep.name, [create_symbol(arg) for arg in rep.arguments]))
    raise RuntimeError("invalid atom: {}".format(rep))

def create_formula(rep, add_formula):
    """
    Returns the temporal formula corresponding the given theory term.

    Throws an error if rep it is not a valid formula.

    Arguments:
    rep         -- Theory term to translate.
    add_formula -- Callback to add resulting formuals.
    """
    if rep.type == _clingo.TheoryTermType.Symbol:
        return create_atom(rep, add_formula, True)
    elif rep.type == _clingo.TheoryTermType.Function:
        args = rep.arguments
        if rep.name in g_binary_operators and len(args) == 2:
            if rep.name in ("|", "&"):
                lhs = create_formula(args[0], add_formula)
                rhs = create_formula(args[1], add_formula)
                return add_formula(TelClause([lhs, rhs], rep.name == "&"))
            else:
                raise RuntimeError("invalid temporal formula: {}".format(rep))
        elif rep.name in g_unary_operators and len(args) == 1:
            arg = create_formula(args[0], add_formula)
            return add_formula(TelNegation(arg))
        elif rep.name in g_tel_operators:
            if rep.name in ("<", "<:", "<;", "<:;", "<*", "<?", "<<"):
                raise RuntimeError("invalid temporal formula: {}".format(rep))
            rhs = create_formula(args[-1], add_formula)
            if rep.name == ">" or rep.name == ">:":
                lhs = 1 if len(args) == 1 else create_number(args[0])
                return rhs if lhs == 0 else add_formula(TelNext(lhs, rhs, rep.name == ">:"))
            lhs = None if len(args) == 1 else create_formula(args[0], add_formula)
            if rep.name in (";>", ";>:"):
                return add_formula(TelClause([lhs, TelNext(1, rhs, rep.name == ";>:")], True))
            elif rep.name in (">*", ">?"):
                return add_formula(TelUntil(lhs, rhs, rep.name == ">?"))
            else:
                assert(rep.name == ">>")
                final = add_formula(TelAtom(True, "__final", []))
                rhs = add_formula(TelClause([add_formula(TelNegation(final)), rhs], False))
                return add_formula(TelUntil(None, rhs, False))
        elif rep.name == "&":
            arg = rep.arguments[0]
            if arg.type == _clingo.TheoryTermType.Symbol:
                if arg.name == "initial" or arg.name == "final":
                    name = "__{}".format(arg.name)
                    return add_formula(TelNegation(TelNegation(TelAtom(True, name, []))))
                elif arg.name == "true" or arg.name == "false":
                    return add_formula(TelConstant(arg.name == "true"))
                else:
                    raise RuntimeError("unknown identifier: {}".format(rep))
            else:
                raise RuntimeError("invalid temporal formula: {}".format(rep))
        else:
            return create_atom(rep, add_formula, True)
    else:
        raise RuntimeError("invalid temporal formula: {}".format(rep))

class ShiftFormula(_tf.TelTransformer):
    """
    Shifts the given formula.
    """
    def __init__(self, shift):
        self.__shift = shift

    def visit_TelAtom(self, x):
        return x if self.__shift == 0 else TelShift(-self.__shift, x)

    def visit_TelNext(self, x):
        if x.lhs <= self.__shift:
            return shift_formula(x.rhs, self.__shift - x.lhs)
        else:
            return TelShift(0, TelNext(x.lhs - self.__shift, x.rhs, x.weak))

    def visit_TelUntil(self, x):
        inner = TelNext(1, x, not x.until)
        if x.lhs is not None:
            inner = TelClause([x.lhs, inner], x.until)
        return shift_formula(TelClause([x.rhs, inner], not x.until), self.__shift)

    def visit_TelClause(self, x):
        return TelClause(self(x.elements), x.conjunctive)

    def visit_TelNegation(self, x):
        return TelShift(-self.__shift, x)

    def visit_TelConstant(self, x):
        return TelShift(-self.__shift, x)

def shift_formula(x, shift):
    return ShiftFormula(shift)(x)

class UnfoldFormula(_tf.TelTransformer):
    """
    Unfolds the given formula into normal rules.
    """
    def visit_TelAtom(self, x):
        return [[x]]

    def visit_TelShift(self, x):
        return [[x]]

    def visit_TelClause(self, x):
        elements = map(self, x.elements)
        # NOTE: list is mapped over the elements to make sure that iterators are consumed only once
        return map(list, _it.chain(*elements) if x.conjunctive else _it.starmap(_it.chain, _it.product(*elements)))

def unfold_formula(x):
    return UnfoldFormula()(x)

class HeadFormulaToBodyFormula(_tf.TelTransformer):
    def __init__(self, add_formula):
        self.__add_formula = add_formula

    def visit_TelAtom(self, x):
        return self.__add_formula(_bd.Atom(x.name, x.arguments, x.positive))

    def visit_TelNext(self, x):
        return self.__add_formula(_bd.Next(self(x.rhs), x.lhs, x.weak))

    def visit_TelUntil(self, x):
        formula = self.__add_formula(_bd.TelFormulaN(">?", None if x.lhs is None else self(x.lhs), self(x.rhs)))
        formula.set_future(self.__add_formula(_bd.Next(formula, 1, False)))
        return formula

    def visit_TelClause(self, x):
        op = "&" if x.conjunctive else "|"
        elements = map(self, x.elements)
        return _ft.reduce(lambda l, r: _bd.BooleanFormula(op, l, r), elements)

    def visit_TelNegation(self, x):
        return self.__add_formula(_bd.Negation(self(x.rhs)))

    def visit_TelConstant(self, x):
        return self.__add_formula(_bd.BooleanConstant(x.value))

def head_formula_to_body_formula(x, add_formula):
    return HeadFormulaToBodyFormula(add_formula)(x)

class ClauseToRule(_tf.TelTransformer):
    def __init__(self, head, body):
        self.__head = head
        self.__body = body

    def visit_TelAtom(self, x, ctx, step):
        sym  = _clingo.Function(x.name, x.arguments + [_clingo.Number(step)], x.positive)
        atom = ctx.symbols[sym]
        if atom is not None:
            self.__head.append(atom.literal)

    def visit_TelShift(self, x, ctx, step):
        stp = lambda x, n, w: x
        if x.lhs != 0:
            stp = _bd.Next if x.lhs > 0 else _bd.Previous
        neg = lambda x: ctx.add_formula(_bd.Negation(x))
        nxt = lambda l, r: ctx.add_formula(stp(r, abs(l), False))
        rhs = head_formula_to_body_formula(x.rhs, ctx.add_formula)
        frm = neg(nxt(x.lhs, rhs))
        self.__body.append(frm.translate(ctx, step))

def translate_clause(clause, ctx, step, body_literal):
    head = []
    body = [body_literal]
    for lit in clause:
        ClauseToRule(head, body)(lit, ctx, step)
    ctx.backend.add_rule(head, body)

class HeadFormula(Formula):
    """
    Class for temporal and Boolean formulas in rule heads.
    """
    def __init__(self, timestep, formula):
        self.__formula = formula
        self.__timestep = timestep
        self.__literals = []

    @property
    def _rep(self):
        """
        Return the unique string representaiton of the formula.
        """
        return ("head", self.__timestep, self.__formula._rep)

    def translate(self, ctx, step):
        """
        Translates a formula at a given step.

        The formula is first shifted (everything not referring to the current
        time step is double negated).  Then the formula can be converted to
        normal rules. This brings it into clausal form, where every negated
        occurrence of a formula is shifted into a rule body.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.

        Possible Future Optimizations:
        - There can be no more derivations if all next operators have been
          unpacked and no inductive temporal operator has been unpacked.
        - Subformulas not referring to a positive atom can be preceded by
          double negation and do not have to be unpacked at all.
        - To make the translations practical, formulas should be factored in a
          way so that become more compact. The current proof-of-concept
          translation does not pay much mind to this.
        """
        shifted = shift_formula(self.__formula, step - self.__timestep)
        undfolded = unfold_formula(shifted)

        if len(self.__literals) > 1:
            body = backend.atom()
            for x in self.__literals:
                backend.add_rule(body, [x])
            self.__literals = [body]

        for clause in undfolded:
            translate_clause(clause, ctx, step, self.__literals[0])

        ctx.add_todo(self, step+1)

    def add_literal(self, literal):
        self.__literals.append(literal)

    def __str__(self):
        return "{}@{}".format(self.__formula, self.__timestep)

    def __repr__(self):
        return "HeadFormula({!r},{!r})".format(self.__timestep, self.__formula)

def translate_formula(atom, add_formula):
    '''
    - add the formula to the above variants
    - print the variants
    '''
    clause = []
    for x in atom.elements:
        if x.condition or len(x.terms) != 1:
            raise RuntimeError('invalid temporal formula: {}'.format(atom))
        clause.append(create_formula(x.terms[0], add_formula))
    formula = HeadFormula(atom.term.arguments[0].number, clause[0] if len(clause) == 1 else TelClause(clause, False))
    formula.add_literal(atom.literal)
    return formula



"""
Module with functions to transform heads referring to the future.
"""

from . import transformer as _tf

import clingo as _clingo
from clingo import ast as _ast
from numbers import Number as _Number
from operator import itemgetter as _itemgetter

# {{{ data structures

g_tel_false_atom = "__false"
g_tel_keywords = ["true", "false", "final", "initial"]
g_tel_shift_variable = "__S"

def time_parameter(loc):
    return _ast.SymbolicTerm(loc, _clingo.Function(_tf.g_time_parameter_name))

class Interval:
    def __init__(self, left, right):
        self.left  = left
        self.right = right

    def before(self, other):
        return self.right < other.left

    def union(self, other):
        self.left  = min(self.left, other.left)
        self.right = max(self.right, other.right)

    def empty(self):
        return self.left >= self.right

    def __repr__(self):
        return "({!r},{!r})".format(self.left, self.right)

    def __str__(self):
        return "[{},{})".format(self.left, self.right)

class IntervalSet:
    def __init__(self, elements = []):
        self.__elements = []
        for x in elements:
            self.add(x)

    def __len__(self):
        return len(self.__elements)

    def add(self, x):
        y = Interval(x[0], x[1])
        if not y.empty():
            i = 0
            while i < len(self) and self.__elements[i].before(y):
                i += 1

            j = i
            while j < len(self) and not y.before(self.__elements[j]):
                y.union(self.__elements[j])
                j += 1

            if i == j:
                self.__elements.insert(i, y)
            else:
                self.__elements[i:j] = (y,)

    def __iter__(self):
        for x in self.__elements:
            yield x.left, x.right

    def __contains__(self, x):
        y = Interval(x[0], x[1])
        if y.empty():
            return True

        i = 0
        while i < len(self) and self.__elements[i].before(y):
            i += 1

        return i < len(self) and self.__elements[i].left <= y.left and y.right <= self.__elements[i].right

    def __repr__(self):
        return "IntervalSet({!r})".format(self.__elements)

    def __str__(self):
        return "{{{}}}".format(",".join((str(x) for x in self.__elements)))

# {{{1 parse_raw_formula

class TheoryParser:
    """
    Parser for temporal formulas.

    Constants:
    unary  -- Boolean to mark unary operators.
    binary -- Boolean to mark unary operators.
    left   -- Boolean to mark left associativity.
    right  -- Boolean to mark right associativity.
    """
    unary, binary = True, False
    left,  right  = True, False
    table = {
        ("&"  , unary):  (7, None),
        ("-"  , unary):  (7, None),
        ("+"  , binary): (6, left),
        ("-"  , binary): (6, left),
        ("~"  , unary):  (5, None),
        (">"  , unary):  (5, None),
        (">"  , binary): (5, right),
        (">:" , unary):  (5, None),
        (">:" , binary): (5, right),
        (">?" , unary):  (5, None),
        (">*" , unary):  (5, None),
        (">>" , unary):  (5, None),
        (">*" , binary): (4, left),
        (">?" , binary): (4, left),
        ("&"  , binary): (3, left),
        ("|"  , binary): (2, left),
        (";>" , binary): (0, right),
        (";>:", binary): (0, right) }

    def __init__(self):
        """
        Initializes the parser.
        """
        self.__stack  = []

    def __priority_and_associativity(self, operator):
        """
        Get priority and associativity of the given binary operator.
        """
        return self.table[(operator, self.binary)]

    def __priority(self, operator, unary):
        """
        Get priority of the given unary or binary operator.
        """
        return self.table[(operator, unary)][0]

    def __check(self, operator):
        """
        Returns true if the stack has to be reduced because of the precedence
        of the given binary operator is lower than the preceeding operator on
        the stack.
        """
        if len(self.__stack) < 2:
            return False
        priority, associativity = self.__priority_and_associativity(operator)
        previous_priority       = self.__priority(*self.__stack[-2])
        return previous_priority > priority or (previous_priority == priority and associativity)

    def __reduce(self):
        """
        Combines the last unary or binary term on the stack.
        """
        b = self.__stack.pop()
        operator, unary = self.__stack.pop()
        if unary:
            self.__stack.append(_ast.TheoryFunction(b.location, operator, [b]))
        else:
            a = self.__stack.pop()
            loc = _ast.Location(a.location.begin, b.location.end)
            self.__stack.append(_ast.TheoryFunction(loc, operator, [a, b]))

    def parse(self, x):
        """
        Parses the given unparsed term, replacing it by nested theory
        functions.
        """
        del self.__stack[:]
        unary = True
        for element in x.elements:
            for operator in element.operators:
                if not (operator, unary) in self.table:
                    raise RuntimeError("invalid operator in temporal formula: {}".format(_tf.str_location(x.location)))
                while not unary and self.__check(operator):
                    self.__reduce()
                self.__stack.append((operator, unary))
                unary = True
            self.__stack.append(element.term)
            unary = False
        while len(self.__stack) > 1:
            self.__reduce()
        return self.__stack[0]

def parse_raw_formula(x):
    """
    Turns the given unparsed term into a term.
    """
    return TheoryParser().parse(x)

# {{{1 theory_term -> term

class TheoryTermToTermTransformer(_ast.Transformer):
    """
    This class transforms a given theory term into a plain term.
    """
    def visit_TheoryTermSequence(self, x):
        """
        Theory term tuples are mapped to term tuples.
        """
        if x.sequence_type == _ast.TheorySequenceType.Tuple:
            return _ast.Function(x.location, "", [self(a) for a in x.arguments], False)
        else:
            raise RuntimeError("invalid term: {}".format(_tf.str_location(x.location)))

    def visit_TheoryFunction(self, x):
        """
        Theory functions are mapped to functions.

        If the function name refers to a temporal operator, an exception is thrown.
        """
        isnum = lambda y: y.ast_type == _ast.ASTType.SymbolicTerm and y.symbol.type == _clingo.SymbolType.Number
        if x.name == "-" and len(x.arguments) == 1:
            rhs = self(x.arguments[0])
            if isnum(rhs):
                return _ast.SymbolicTerm(x.location, _clingo.Number(-rhs.symbol.number))
            else:
                return _ast.UnaryOperation(x.location, _ast.UnaryOperator.Minus, rhs)
        elif (x.name == "+" or x.name == "-") and len(x.arguments) == 2:
            lhs = self(x.arguments[0])
            rhs = self(x.arguments[1])
            op  = _ast.BinaryOperator.Plus if x.name == "+" else _ast.BinaryOperator.Minus
            if isnum(lhs) and isnum(rhs):
                lhs = lhs.symbol.number
                rhs = rhs.symbol.number
                return _ast.SymbolicTerm(x.location, _clingo.Number(lhs + rhs if x.name == "+" else lhs - rhs))
            else:
                return _ast.BinaryOperation(x.location, op, lhs, rhs)
        elif x.name == "-" and len(x.arguments) == 2:
            return _ast.BinaryOperation(x.location, _ast.BinaryOperator.Minus, self(x.arguments[0]), self(x.arguments[1]))
        elif (x.name, TheoryParser.binary) in TheoryParser.table or (x.name, TheoryParser.unary) in TheoryParser.table:
            raise RuntimeError("invalid term: {}".format(_tf.str_location(x.location)))
        else:
            return _ast.Function(x.location, x.name, [self(a) for a in x.arguments], False)

    def visit_TheoryUnparsedTerm(self, x):
        """
        Unparsed term are first parsed and then handled by the transformer.
        """
        return self.visit(parse_raw_formula(x))

def theory_term_to_term(x):
    """
    Convert the given theory term into a term.
    """
    return TheoryTermToTermTransformer()(x)

# {{{1 theory_term -> symbolic_atom

class TheoryTermToAtomTransformer(_ast.Transformer):
    """
    Turns the given theory term into an atom.
    """

    def __atom(self, location, positive, name, arguments):
        """
        Helper function to create an atom.

        Arguments:
        location --  Location to use.
        positive --  Classical sign of the atom.
        name     --  The name of the atom.
        arguments -- The arguments of the atom.
        """
        ret = _ast.Function(location, name, arguments + [time_parameter(location)], False)
        if not positive:
            ret = _ast.UnaryOperation(location, _ast.UnaryOperator.Minus, ret)
        return _ast.SymbolicAtom(ret)

    def visit_SymbolicTerm(self, x, positive):
        """
        Maps functions to atoms.

        Every other symbol causes a runtime error.

        Arguments:
        x        -- The theory term to translate.
        positive -- The classical sign of the atom.
        """
        symbol = x.symbol
        if x.symbol.type == _clingo.SymbolType.Function and len(symbol.name) > 0:
            return self.__atom(x.location, positive == symbol.positive, symbol.name, [_ast.Symbol(x.location, a) for a in symbol.arguments])
        else:
            raise RuntimeError("invalid temporal formula in rule head: {}".format(_tf.str_location(x.location)))

    def visit_Variable(self, x, positive):
        """
        Raises an error.
        """
        raise RuntimeError("invalid temporal formula in rule head: {}".format(_tf.str_location(x.location)))


    def visit_TheoryTermSequence(self, x, positive):
        """
        Raises an error.
        """
        raise RuntimeError("invalid temporal formula in rule head: {}".format(_tf.str_location(x.location)))

    def visit_TheoryFunction(self, x, positive):
        """
        Maps theory functions to atoms.

        If the function name refers to a temporal operator, an exception is thrown.
        """
        if x.name == "-":
            return self(x.arguments[0], not positive)
        elif (x.name, TheoryParser.binary) in TheoryParser.table or (x.name, TheoryParser.unary) in TheoryParser.table:
            raise RuntimeError("invalid term: {}".format(_tf.str_location(x.location)))
        else:
            return self.__atom(x.location, positive, x.name, [theory_term_to_term(a) for a in x.arguments])

    def visit_TheoryUnparsedTerm(self, x, positive):
        """
        Unparsed terms are first parsed and then handled by the transformer.
        """
        return self.visit(parse_raw_formula(x), positive)

def theory_term_to_atom(x, positive=True):
    """
    Convert the given theory term into an atom.
    """
    return TheoryTermToAtomTransformer()(x, positive)

# {{{1 theory transformers

class TheoryAtomTransformer(_ast.Transformer):
    """
    Transforms the given theory atom to be processed further.
    """

    def __init__(self, atoms):
        self.__atoms = atoms

    def __add_atom(self, x, rng):
        self.__atoms.append((theory_term_to_atom(x), rng))

    def __add_range(self, location, rng, left, right):
        def add(a, b):
            if a == 0:
                return b
            elif b == 0:
                return a
            elif a == float("inf") or b == float("inf"):
                return float("inf")
            elif isinstance(a, _Number) and isinstance(b, _Number):
                return a + b
            else:
                lhs = _ast.SymbolicTerm(location, _clingo.Number(a)) if isinstance(a, _Number) else a
                rhs = _ast.SymbolicTerm(location, _clingo.Number(b)) if isinstance(b, _Number) else b
                return _clingo.ast.BinaryOperation(location, _ast.BinaryOperator.Plus, lhs, rhs)

        return add(left, rng[0]), add(right, rng[1])

    def visit_SymbolicTerm(self, x, rng):
        self.__add_atom(x, rng)
        return x

    def visit_Variable(self, x, rng):
        """
        Raises an error.
        """
        raise RuntimeError("invalid temporal formula in rule head: {}".format(_tf.str_location(x.location)))

    def visit_TheoryTermSequence(self, x, rng):
        """
        Raises an error.
        """
        raise RuntimeError("invalid temporal formula in rule head: {}".format(_tf.str_location(x.location)))

    def visit_TheoryUnparsedTerm(self, x, rng):
        """
        Unparsed terms are first parsed and then handled by the transformer.
        """
        return self.visit(parse_raw_formula(x), rng)

    def visit_TheoryFunction(self, x, rng):
        """
        Transforms the given theory function into a temporal formula.
        """
        is_binary = (x.name, TheoryParser.binary) in TheoryParser.table and len(x.arguments) == 2
        is_unary  = (x.name, TheoryParser.unary) in TheoryParser.table and len(x.arguments) == 1
        if is_unary or is_binary:
            if x.name == "-":
                self.__add_atom(x, rng)
                return x
            elif x.name == "~":
                return x

            lhs = None if is_unary else x.arguments[0]
            rhs = x.arguments[0 if is_unary else 1]

            if x.name == ">" or x.name == ">:":
                if lhs is None:
                    lhs = 1
                else:
                    lhs = theory_term_to_term(x.arguments[0])
                    if lhs.ast_type == _ast.ASTType.SymbolicTerm and lhs.symbol.type == _clingo.SymbolType.Number:
                        lhs = lhs.symbol.number
                self(rhs, self.__add_range(x.location, rng, lhs, lhs))
            elif x.name == "&" and lhs is None:
                if rhs.ast_type != _ast.ASTType.SymbolicTerm or len(rhs.symbol.arguments) != 0 or rhs.symbol.name not in g_tel_keywords:
                    raise RuntimeError("invalid temporal formula in rule head: {}".format(_tf.str_location(x.location)))
            else:
                rng_left, rng_right = rng, rng
                if x.name == ">?" or x.name == ">*" or x.name == ">>":
                    rng_left = self.__add_range(x.location, rng, 0, float("inf"))
                    rng_right = rng_left
                elif x.name == ";>" or x.name == ";>:":
                    rng_right = self.__add_range(x.location, rng, 1, 1)
                if is_binary:
                    self(lhs, rng_left)
                self(rhs, rng_right)
        else:
            self.__add_atom(x, rng)

        return x

    def visit_TheoryAtomElement(self, x):
        """
        Transforms one elementary theory elements without conditions into formulas.
        """
        # NOTE: in principle this condition can be relaxed...
        if len(x.terms) != 1 or len(x.condition) != 0:
            raise RuntimeError("invalid temporal formula in rule head: {}".format(_tf.str_location(x.location)))
        x.terms[0] = self(x.terms[0], (0, 0))
        return x

    def visit_TheoryAtom(self, x):
        """
        Transforms the given theory atom to be processed further.

        The theory atom is renamed from tel to tel_head(__t) and the
        """
        if x.guard is not None:
            raise RuntimeError("invalid temporal formula in rule head: {}".format(_tf.str_location(x.location)))
        x.term     = _ast.Function(x.term.location, "__tel_head", [time_parameter(x.term.location)], False)
        x.elements = [self(elem) for elem in x.elements]
        return x

def transform_theory_atom(x):
    """
    Transforms the given theory atom.
    """
    atoms = []
    atom = TheoryAtomTransformer(atoms)(x)

    # maps atoms to a set of numeric ranges
    numeric = {}
    # maps ranges to a set of symbolic ranges
    other = {}

    # add to symbolic ranges converting numeric ranges
    def add(atm, lhs, rhs):
        if isinstance(lhs, _Number):
            lhs = _ast.SymbolicTerm(atom.location, _clingo.Number(lhs))
        if rhs == float("inf"):
            rhs = _ast.SymbolicTerm(atom.location, _clingo.Supremum)
        elif isinstance(rhs, _Number):
            rhs = _ast.SymbolicTerm(atom.location, _clingo.Number(rhs))

        rng = (lhs, rhs)
        other.setdefault(rng, (rng, {}))[1].setdefault(atm, atm)

    # split into numeric and symbolic ranges
    for atm, (lhs, rhs) in atoms:
        if isinstance(lhs, _Number) and isinstance(rhs, _Number):
            numeric.setdefault(atm, (atm, IntervalSet()))[1].add((lhs, rhs+1))
        else:
            add(atm, lhs, rhs)

    # add combined numeric ranges as symbolic ranges
    for atm, rngs in numeric.values():
        for lhs, rhs in rngs:
            add(atm, lhs, rhs-1)

    # flatten symbolic ranges into a list
    ranges = []
    for _, (rng, atoms) in sorted(other.items(), key=_itemgetter(0)):
        ranges.append((rng, [atm for _, atm in sorted(atoms.items(), key=_itemgetter(0))]))

    return atom, ranges

# {{{1 get_variables

class VariablesVisitor(_ast.Transformer):
    """
    Visitor to collect variables.

    Attributes:
    __variables -- reference to the resulting list of variables
    """
    def __init__(self, variables):
        """
        Initializes the visitor with a reference to a list for storing the
        visited variables.
        """
        self.__variables = variables

    def visit_Variable(self, x):
        """
        Stores the variable in the list.
        """
        self.__variables[str(x)] = x
        return x

def get_variables(x):
    """
    Gets all variables in a formula.
    """
    v = {}
    VariablesVisitor(v)(x)
    return [val for _, val in sorted(v.items(), key=lambda x: x[0])]

# {{{1 transform_head

# TODO: Something like this will be needed later at another place...
def factor_out_tel_formula(x):
    if x.type == "TelClause":
        if x.conjunctive:
            ret = []
            for y in x.elements:
                ret.extend(factor_out_tel_formula(y))
        else:
            ret = factor_out_tel_formula(x.elements[0])
            for y in x.elements[1:]:
                ret = [a + b for b in factor_out_tel_formula(y) for a in ret]
        return ret
    else:
        return [[x]]

class HeadTransformer:
    def __init__(self):
        self.__num_aux = 0
        self.__false_external = None

    def __aux_atom(self, location, variables, inc=1):
        self.__num_aux += inc
        return _ast.Literal(location, _ast.Sign.NoSign, _ast.SymbolicAtom(_ast.Function(location, "__aux_{}".format(self.__num_aux - 1), variables, False)))

    def __false_atom(self, location):
        return _ast.Literal(location, _ast.Sign.NoSign, _ast.SymbolicAtom(_ast.Function(location, g_tel_false_atom, [time_parameter(location)], False)))

    def transform(self, atom):
        loc          = atom.location
        false        = self.__false_atom(loc)
        atom, ranges = transform_theory_atom(atom)
        variables    = get_variables(atom)
        param        = time_parameter(loc)
        shift        = _ast.Variable(loc, g_tel_shift_variable)
        aux          = self.__aux_atom(loc, variables + [param])
        saux         = self.__aux_atom(loc, variables + [shift], inc=0)
        rules        = []

        if self.__false_external is None:
            rules.append(_tf.External(loc, false.atom, []))

        rules.append(_ast.Rule(loc, atom, [aux]))

        if ranges:
            elems = []
            for (lhs, rhs), heads in ranges:
                cond = []
                diff = _ast.BinaryOperation(loc, _ast.BinaryOperator.Minus, param, shift)
                if lhs.ast_type != _ast.ASTType.SymbolicTerm or lhs.symbol.type != _clingo.SymbolType.Number or lhs.symbol.number > 0:
                    cond.append(_ast.Literal(loc, _ast.Sign.NoSign, _ast.Comparison(_ast.ComparisonOperator.LessEqual, lhs, diff)))

                if rhs.ast_type != _ast.ASTType.SymbolicTerm or rhs.symbol.type != _clingo.SymbolType.Supremum:
                    cond.append(_ast.Literal(loc, _ast.Sign.NoSign, _ast.Comparison(_ast.ComparisonOperator.LessEqual, diff, rhs)))

                elems.extend([_ast.ConditionalLiteral(loc, _ast.Literal(loc, _ast.Sign.NoSign, head), cond) for head in heads])

            rules.append(_ast.Rule(loc, _ast.Disjunction(loc, elems), [saux, false]))

        return aux, rules



"""
This module contains the basic transformer to traves and change an AST as well
as constants used during translation.

Classes:
Transformer -- Base class to modify ASTs.

Functions:
str_location   -- Turn a location into a string.
is_constraint  -- Check whether a statement is a constraint.
is_normal      -- Check whether a statement is a normal rule.
is_disjunction -- Check whether a statement is a disjunctive rule.

Constants:
g_future_prefix           -- Prefix for predicates referring to the future.
g_variable_prefix         -- Prefix for auxiliary variables.
g_time_parameter_name     -- Prefix for the time parameter.
g_time_parameter_name_alt -- Prefix for the second time parameter used when
                             grounding rules within a given range.
"""

import clingo as _clingo
from clingo import ast as _ast

g_future_prefix = "__future_"
g_variable_prefix = "X"
g_time_parameter_name = "__t"
g_time_parameter_name_alt = "__u"

def str_location(loc):
    """
    This function takes a location from a clingo AST and transforms it into a
    readable format.
    """
    begin = loc.begin
    end   = loc.end
    ret = "{}:{}:{}".format(begin.filename, begin.line, begin.column)
    dash = True
    eq = begin.filename == end.filename
    if not eq:
        ret += "{}{}".format("-" if dash else ":", end.filename)
        dash = False
    eq = eq and begin.line == end.line
    if not eq:
        ret += "{}{}".format("-" if dash else ":", end.line)
        dash = False
    eq = eq and begin.column == end.column
    if not eq:
        ret += "{}{}".format("-" if dash else ":", end.column)
        dash = False
    return ret

def is_constraint(s):
    """
    Check if the given AST node is a constraint.

    As a special case this function also considers rules with a negative
    literal in the head as a constraint.
    """
    return (s.ast_type == _ast.ASTType.Rule and s.head.ast_type == _ast.ASTType.Literal and
            ((s.head.atom.ast_type == _ast.ASTType.BooleanConstant and not s.head.atom.value) or
             (s.head.sign != _ast.Sign.NoSign)))

def is_normal(s):
    """
    Check if the given statement is a normal rule.
    """
    return (s.ast_type == _ast.ASTType.Rule and
            s.head.ast_type == _ast.ASTType.Literal and
            s.head.sign == _ast.Sign.NoSign and
            s.head.atom.ast_type == _ast.ASTType.SymbolicAtom)

def is_disjunction(s):
    """
    Check if a given AST node is a disjunction.

    Normal rules and constraints are not conisdered disjunctions.
    """
    return (s.ast_type == _ast.ASTType.Rule and s.head.ast_type == _ast.ASTType.Disjunction)

class TelTransformer:
    """
    Basic visitor to traverse and modify an AST.

    Transformers to modify an AST should subclass this class and add visit_TYPE
    methods where TYPE corresponds to an ASTType. This function is called
    whenever a node of the respective type is visited. Its return value will
    replace the node in the parent.

    Function visit should be called on the root of the AST to be visited. It is
    the users responsibility to visit children of nodes that have node-specific
    visitor.
    """
    def visit_children(self, x, *args, **kwargs):
        """
        Visits and transforms the children of the given node.
        """
        updated = []
        for key in x.keys:
            if key in x.child_keys:
                value = self.visit(getattr(x, key), *args, **kwargs)
                updated.append(value)
            else:
                updated.append(getattr(x, key))
        return x.__class__(*updated)

    def visit(self, x, *args, **kwargs):
        """
        Visits the given node and returns its transformation.

        If there is a matching visit_TYPE function where TYPE corresponds to
        the ASTType of the given node then this function called and its value
        returned. Otherwise, its children are visited and transformed.

        This function accepts additional positional and keyword arguments,
        which are passed to node-specific visit functions and to the visit
        function called for child nodes.
        """
        if hasattr(x, "ast_type"):
            attr = "visit_" + str(x.ast_type)
            if hasattr(self, attr):
                return getattr(self, attr)(x, *args, **kwargs)
            else:
                return self.visit_children(x, *args, **kwargs)
        elif isinstance(x, list):
            return [self.visit(y, *args, **kwargs) for y in x]
        elif x is None:
            return x
        else:
            raise TypeError("unexpected type")

    def __call__(self, x, *args, **kwargs):
        """
        Alternative way to call visit.
        """
        return self.visit(x, *args, **kwargs)

_version = _clingo.__version__.split(".")
if int(_version[0]) >= 5 and int(_version[1]) >= 4:
    External = lambda loc, head, body: _ast.External(loc, head, body, _ast.Function(loc, "false", [], False))
else:
    External = _ast.External



"""
This module is responsible for visiting and rewriting a programs AST to account
for temporal operators.

Handling of normal atoms
========================
The temporal program

  p :- q.

becomes

  p(t) :- q(t).

Handling of past atoms in the body
==================================
The temporal program

  p :- 'q.

becomes

  p(t) :- q(t-1).

Handling of future atoms in the head
====================================
The temporal program

  p'.

referring to the future in a normal rule head is rewritten into the following
ASP program:

  f_p(1,t+1).

with auxiliary rules

  #program always(t).
  p(t) :- f_p(1,t).

and future signatures [('f_p', 2, True)] whose atoms have to be set to False if
referring to the future.

Handling of constraints referring to the future
===============================================
The temporal program

  :- p''.

is rewritten into

  #program always(t).
  #program always_0_1(t,u).
  :- p((t+2)); __final(u).
  #program always_2(t,u).
  :- p((t+2)).

where the constraint is removed from the always part and put into two
additional program parts which have to be grounded for time points t, t-1, and
t-2 as given by the last return value:

  [('always', 'always_0_1', range(0, 2)),
   ('always', 'always_2',   range(2, 3))]

Functions:
transform -- transforms telingo programs into incremental ASP
"""

from . import transformer as _tf
from . import program as _prg

import clingo as _clingo
from clingo import ast as _ast
from textwrap import dedent as _dedent

def transform(inputs, callback):
    """
    Transforms the given list of temporal programs in string form into an ASP
    program.

    Returns the future predicates whose atoms have to be set to false if
    referring to the future, and program parts that have to be regrounded if
    there are constraints referring to the future.

    Arguments:
    inputs   -- The list of inputs.
    callback -- Callback for rewritten statements.
    """
    loc = _ast.Location(_ast.Position('<transform>', 1, 1), _ast.Position('<transform>', 1, 1))
    future_predicates = set()
    constraint_parts  = {}
    time              = _ast.SymbolicTerm(loc, _clingo.Function(_tf.g_time_parameter_name))
    wrap_lit          = lambda a: _ast.Literal(loc, _ast.Sign.NoSign, a)

    # apply transformer to program
    def append(s):
        if s is not None:
            callback(s)
    aux_rules = []
    transformer = _prg.ProgramTransformer(future_predicates, constraint_parts, aux_rules)
    for i in inputs:
        _ast.parse_string(i, lambda s: append(transformer.visit(s)))
    if aux_rules:
        callback(_ast.Program(loc, "always", [_ast.Id(loc, _tf.g_time_parameter_name), _ast.Id(loc, _tf.g_time_parameter_name_alt)]))
        for rule in aux_rules:
            callback(rule)

    # add auxiliary rules for future predicates
    future_sigs = []
    if len(future_predicates) > 0:
        callback(_ast.Program(loc, "always", [_ast.Id(loc, _tf.g_time_parameter_name), _ast.Id(loc, _tf.g_time_parameter_name_alt)]))
        for name, arity, positive, shift in sorted(future_predicates):
            variables = [ _ast.Variable(loc, "{}{}".format(_tf.g_variable_prefix, i)) for i in range(arity) ]
            s = _ast.SymbolicTerm(loc, _clingo.Number(shift))
            t_shifted = _ast.BinaryOperation(loc, _ast.BinaryOperator.Plus, time, s)
            add_sign = lambda lit: lit if positive else _ast.UnaryOperation(loc, _ast.UnaryOperator.Minus, lit)
            p_current = _ast.SymbolicAtom(add_sign(_ast.Function(loc, name, variables + [time], False)))
            f_current =  _ast.SymbolicAtom(add_sign(_ast.Function(loc, _tf.g_future_prefix + name, variables + [s, time], False)))
            callback(_ast.Rule(loc, wrap_lit(p_current), [wrap_lit(f_current)]))
            future_sigs.append((_tf.g_future_prefix + name, arity + 2, positive))

    # gather rules for constraints referring to the future
    reground_parts = []
    if len(constraint_parts) > 0:
        for (name, shift), rules in constraint_parts.items():
            assert(shift > 0)
            params = [_ast.Id(loc, _tf.g_time_parameter_name), _ast.Id(loc, _tf.g_time_parameter_name_alt)]
            # parts to be regrounded
            part = "{}_0_{}".format(name, shift-1)
            callback(_ast.Program(loc, part, params))
            for p, l in rules:
                callback(p)
            reground_parts.append((name, part, range(shift)))
            # parts that no longer have to be regrounded
            last_part = "{}_{}".format(name, shift)
            callback(_ast.Program(loc, last_part, params))
            for p, l in rules:
                callback(l)
            reground_parts.append((name, last_part, range(shift, shift+1)))

    def add_part(part_name, atom_name, statement, wrap=lambda x: x):
        params = [_ast.Id(loc, _tf.g_time_parameter_name), _ast.Id(loc, _tf.g_time_parameter_name_alt)]
        callback(_ast.Program(loc, part_name, params))
        atom = wrap(_ast.SymbolicAtom(_ast.Function(loc, atom_name, [time], False)))
        callback(statement(loc, atom, []))
    add_part('initial', '__initial', _ast.Rule, wrap_lit)
    add_part('always', '__final', _tf.External)

    reground_parts.append(('always',  'always',  range(1)))
    reground_parts.append(('dynamic', 'dynamic', range(1)))
    reground_parts.append(('initial', 'initial', range(1)))

    def no_program(s):
        if s.ast_type != _ast.ASTType.Program:
            callback(s)

    _ast.parse_string(_dedent('''\
        #theory tel {
            formula_body  {
                &   : 7, unary;         % prefix for keywords
                -   : 7, unary;         % classical negation
                +   : 6, binary, left;  % arithmetic +
                -   : 6, binary, left;  % arithmetic -
                ~   : 5, unary;         % negation
                <   : 5, unary;         % previous
                <   : 5, binary, right; % n x previous
                <:  : 5, unary;         % weak previous
                <:  : 5, binary, right; % n x weak previous
                <?  : 5, unary;         % eventually-
                <*  : 5, unary;         % always-
                <<  : 5, unary;         % initially
                >   : 5, unary;         % next
                >   : 5, binary, right; % n x next
                >:  : 5, unary;         % weak next
                >:  : 5, binary, right; % n x weak next
                >?  : 5, unary;         % eventually+
                >*  : 5, unary;         % always+
                >>  : 5, unary;         % finally
                >*  : 4, binary, left;  % release
                >?  : 4, binary, left;  % until
                <*  : 4, binary, left;  % trigger
                <?  : 4, binary, left;  % since
                &   : 3, binary, left;  % and
                |   : 2, binary, left;  % or
                <-  : 1, binary, left;  % left implication
                ->  : 1, binary, left;  % right implication
                <>  : 1, binary, left;  % equivalence
                ;>  : 0, binary, right; % sequence next
                ;>: : 0, binary, right; % sequence weak next
                <;  : 0, binary, left;  % sequence previous
                <:; : 0, binary, left   % sequence weak previous
            };
            formula_head  {
                &   : 7, unary;         % prefix for keywords
                -   : 7, unary;         % classical negation
                +   : 6, binary, left;  % arithmetic +
                -   : 6, binary, left;  % arithmetic -
                ~   : 5, unary;         % negation
                >   : 5, unary;         % next
                >   : 5, binary, right; % n x next
                >:  : 5, unary;         % weak next
                >:  : 5, binary, right; % n x weak next
                >?  : 5, unary;         % eventually+
                >*  : 5, unary;         % always+
                >>  : 5, unary;         % finally
                >*  : 4, binary, left;  % release
                >?  : 4, binary, left;  % until
                &   : 3, binary, left;  % and
                |   : 2, binary, left;  % or
                ;>  : 0, binary, right; % sequence next
                ;>: : 0, binary, right  % sequence weak next
            };
            &tel/1 : formula_body, body;
            &__tel_head/1 : formula_body, head
        }.
        '''), no_program)


    _ast.parse_string(_dedent('''\
        #theory del {
            formula_body  {
                &   : 7, unary;         % prefix for keywords
                ?   : 4, unary;         % check
                *   : 3, unary;         % kleene star
                +   : 2, binary, left;  % choice
                ;;  : 1, binary, left;  % sequence
                .>? : 0, binary, right; % diamond (eventually)
                .>* : 0, binary, right  % box (always)
            };
            &del/1 : formula_body, body
        }.
        '''), no_program)

    return future_sigs, reground_parts



"""
Module with functions to transform terms.

Classes:
TermTransformer -- Class to transform terms.
"""


import clingo as _clingo
from clingo import ast as _ast

from . import transformer as _tf

class TermTransformer(_ast.Transformer):
    """
    This class traverses the AST of a term until a Function is found. It then
    add a time parameter to its argument and optionally rewrites the and
    records the predicate name.

    Members:
    parameter         -- Time parameter to extend atoms with.
    future_predicates -- Reference to the set of future predicates
                         having type '{(name, arity, positive, shift)}'
                         where shift corresponds to the number of next
                         operators and positive whether the literal is
                         positive.
    """
    def __init__(self, future_predicates):
        """
        Parameters:
        future_predicates -- reference to the map of future predicates
        """
        self.__future_predicates = future_predicates
        self.__positive = True

    def __get_param(self, name, arity, location, replace_future, fail_future, fail_past, max_shift):
        """
        Strips previous and next operators from function names
        and returns the updated name plus the time arguments to append.
        Furthermore, if the initially operator (_ prefix) is used, then the
        time parameter is replaced with 0. Otherwise, it is treated like a past
        operator.

        If replace_future is set this also introduces a new name for the
        predicate, which is recorded in the list of atoms that have to be made
        redefinable. In this case the name is prefixed with __future_. Such
        dynamic predicates are recorded in the future_predicates list.

        Arguments:
        name           -- The name of the predicate
                          (trailing primes denote previous operators).
        location       -- Location for generated terms.
        replace_future -- Whether atoms referring to the future have to be
                          replaced by a special future atom.
        fail_future    -- Fail if the atom refers to the future.
        fail_past      -- Fail if the atom refers to the past.
        max_shift      -- The maximum number of steps terms look into the
                          future.

        Example for body atoms:

            p(X) :- 'q(X)

        becomes

            p(X,t) :- q(X,t-1)

        Example for head atoms (replace_future=True):

            p''(X) :- q(X).

        becomes

            __future__p(X,2,t) :- q(X,t).

        and future_predicates is extended with (p,1,2) -> False
        """
        n = name.strip("'")
        shift = 0
        for c in name:
            if c == "'":
                shift -= 1
            else:
                break
        shift += len(name) - len(n) + shift

        initially = False
        if n.startswith("_") and not n.startswith("__"):
            n = n[1:]
            if n.startswith("'") or name.startswith("'") or name.endswith("'"):
                raise RuntimeError("initially operator cannot be used with primes: {}".format(_tf.str_location(location)))
            initially = True

        finally_ = False
        if n.endswith("_") and not n.endswith("__"):
            n = n[:-1]
            if n.endswith("'") or name.startswith("'") or name.endswith("'"):
                raise RuntimeError("finally operator cannot be used with primes: {}".format(_tf.str_location(location)))
            finally_ = True
            raise RuntimeError("finally operator not yet supported: {}".format(_tf.str_location(location)))

        if initially and finally_:
            raise RuntimeError("finally and initially operator cannot used together: {}".format(_tf.str_location(location)))
        params = [_ast.SymbolicTerm(location, _clingo.Function(_tf.g_time_parameter_name))]
        if fail_future and (shift > 0 or finally_):
            raise RuntimeError("future atoms not supported in this context: {}".format(_tf.str_location(location)))
        if fail_past and (shift < 0 or initially):
            raise RuntimeError("past atoms not supported in this context: {}".format(_tf.str_location(location)))
        if shift > 0:
            if replace_future:
                self.__future_predicates.add((n, arity, self.__positive, shift))
                n = _tf.g_future_prefix + n
                params.insert(0, _ast.SymbolicTerm(location, _clingo.Number(shift)))
            else:
                max_shift[0] = max(max_shift[0], shift)
        if shift != 0:
            params[-1] = _ast.BinaryOperation(location, _ast.BinaryOperator.Plus, params[-1], _ast.SymbolicTerm(location, _clingo.Number(shift)))
        elif initially:
            params[-1] = _ast.SymbolicTerm(location, _clingo.Number(0))
        return (n, params)

    def visit_UnaryOperation(self, term, *args, **kwargs):
        """
        Determins classical negation.
        """
        try:
            self.__positive = not self.__positive
            return term.update(**self.visit_children(term, *args, **kwargs))
        finally:
            self.__positive = not self.__positive
        return term

    def visit_Function(self, term, *args, **kwargs):
        """
        Transforms the given term.

        See __get_param for more information.
        """
        term.name, params = self.__get_param(term.name, len(term.arguments), term.location, *args, **kwargs)
        term.arguments.extend(params)
        return term

    def visit_SymbolicTerm(self, term, *args, **kwargs):
        """
        Raises a runtime error.

        This function is not necessary if gringo's parser is used but this case
        could occur in a valid AST.
        """
        raise RuntimeError("not implemented")



"""
Module holding basic functions and classes used in the translation temporal
formulas in rule heads and bodies.
"""

import abc as _abc
import clingo as _clingo

def make_equal(backend, a, b):
    """
    Generates clauses for a <-> b.

    Arguments:
    backend -- Backend to add clauses to.
    a       -- first literal
    b       -- second literal
    """
    backend.add_rule([], [ a, -b])
    backend.add_rule([], [-a,  b])

def make_disjunction(backend, e, a, b):
    """
    Generates clauses for e <-> a | b.

    Arguments:
    backend -- Backend to add clauses to.
    e       -- equivalent literal
    a       -- first literal of disjunction
    b       -- second literal of disjunction
    """
    backend.add_rule([], [ e, -a, -b])
    backend.add_rule([], [-e, a])
    backend.add_rule([], [-e, b])

class Context:
    """
    Class gathering arguments used throughout functions in this module.

    Members:
    add_todo        -- Function to add theory atoms that have to be translated
                       later.
    backend         -- Clingo Backend object.
    symbols         -- Clingo SymbolicAtoms object.
    horizon         -- Current search horizon.
    __false_literal -- Function to obtain a false literal.
    """
    def __init__(self, backend, symbols, add_todo, add_formula, false_literal, horizon):
        """
        Initializes the context.

        Arguments:
        backend       -- Backend object.
        symbols       -- SymbolicAtoms object.
        add_todo      -- Function to add theory atoms to the todo list.
        false_literal -- Function to obtain a false literal.
        """
        self.add_todo        = add_todo
        self.add_formula     = add_formula
        self.backend         = backend
        self.symbols         = symbols
        self.horizon         = horizon
        self.__false_literal = false_literal

    @property
    def false_literal(self):
        """
        Returns a literal that is always false.
        """
        return self.__false_literal(self.backend)

"""
Map from binary Boolean connective strings to their ids.
"""
g_binary_operators = {"&", "|", "<>", "<-", "->"}

"""
Set of unary Boolean operators.
"""
g_unary_operators = {"~"}

"""
Set of binary arithmetic operators.
"""
g_arithmetic_operators = {"+", "-"}

"""
Set of temporal connectives.
"""
g_tel_operators = {"<", ">", "<:", ">:", "<*", ">*", ">?", "<?", ">>", "<<", "<;", "<:;", ";>", ";>:"}

"""
Set of dynamic connectives.
"""
g_del_operators = {".>*", ".>?"}

"""
Set of unary Path operators.

"""
g_path_unary_operators  = {"*", "?"}

"""
Set of binary Path operators.

"""
g_path_binary_operators = {";;", "+"}

"""
Set of all tel, del and path operators.

"""
g_all_operators = g_binary_operators.union(g_unary_operators, g_arithmetic_operators, g_tel_operators, g_del_operators, g_path_unary_operators, g_path_binary_operators)

class Formula:
    """
    Base class of all temporal and Boolean formulas.
    """
    @_abc.abstractproperty
    def _rep(self):
        """
        Return the unique string representaiton of the formula.
        """
        pass

    @_abc.abstractmethod
    def translate(self, ctx, step):
        """
        Translates a formula at a given step.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        """
        pass

def create_number(rep):
    if rep.type == _clingo.TheoryTermType.Function:
        args = rep.arguments
        if rep.name == "-" and len(args) == 1:
            return -create_number(rep.arguments[0])
        if rep.name in g_arithmetic_operators and len(args) == 2:
            lhs = create_number(rep.arguments[0])
            rhs = create_number(rep.arguments[1])
            if rep.name == "+":
                return lhs + rhs
            elif rep.name == "-":
                return lhs - rhs
    elif rep.type == _clingo.TheoryTermType.Number and rep.number >= 0:
        return rep.number
    # TODO: this case should be handled as in AG
    #       the corresponding formula should evaluate to false
    raise RuntimeError("number expected: {}".format(rep))

def create_symbol(rep):
    """
    Returns the symbolic representation of the given theory term.

    Throws an error if rep it is not a valid symbol.

    Arguments:
    rep -- Theory term to translate.
    """
    if rep.type == _clingo.TheoryTermType.Number:
        return _clingo.Number(rep.number)
    elif rep.type in [_clingo.TheoryTermType.List, _clingo.TheoryTermType.Set]:
        raise RuntimeError("invalid symbol: {}".format(rep))
    elif rep.type == _clingo.TheoryTermType.Function and rep.name in g_arithmetic_operators:
        if len(rep.arguments) == 1:
            rhs = create_symbol(rep.arguments[0])
            if rep.name == "-":
                if rhs.type == _clingo.SymbolType.Number:
                    return _clingo.Number(-rhs.number)
                elif rhs.type == _clingo.SymbolType.Function and len(rep.name) > 0:
                    return _clingo.Function(rhs.name, rhs.arguments, not rhs.positive)
        elif len(rep.arguments) == 2:
            return _clingo.Number(create_number(rep))

        raise RuntimeError("invalid symbol: {}".format(rep))
    else:
        name = "" if rep.type == _clingo.TheoryTermType.Tuple else rep.name
        if name in g_binary_operators or name in g_unary_operators or name in g_tel_operators:
            raise RuntimeError("invalid symbol: {}".format(rep))
        args = [] if rep.type == _clingo.TheoryTermType.Symbol else rep.arguments
        if len(args) == 0:
            if name == "#inf":
                return _clingo.Infimum
            elif name == "#sup":
                return _clingo.Supremum
            elif len(name) > 1 and name.startswith('"') and name.endswith('"'):
                return _clingo.String(name[1:-1])
        return _clingo.Function(name, [create_symbol(arg) for arg in args])




"""
This module contains simple helper functions.
"""

def getattr_(obj, *attrs):
    """
    Returns the first attribute of the given object that is available.
    """
    for attr in attrs:
        if hasattr(obj, attr):
            return getattr(obj, attr)
    raise AttributeError(
        '{} does not have any of the attributes {}'.format(
            obj.__class__.name,
            ','.join(('"{}"'.format(attr) for attr in attrs))))



#!/usr/bin/env python

from setuptools import setup

setup(
    version = '2.1.1',
    name = 'telingo',
    description = 'System to solve dynamic temporal logic programs.',
    author = 'Roland Kaminski & Francois Laferriere',
    license = 'MIT',
    packages = ['telingo', 'telingo.theory', 'telingo.transformers'],
    test_suite = 'telingo.tests',
    zip_safe = False,
    entry_points = {
        'console_scripts': [
            'telingo = telingo:main',
        ]
    }
)




"""
This module exports functions to translate (ground) theory atoms occuring in
rule bodies to rules via clingo's backend.
"""

import clingo as _clingo
from telingo.util import getattr_

from .formula import *
from .path import *

# Base for Body Formulas {{{1

class StepData:
    """
    Stores step specific inforation of theory atoms, like the set of literals
    it is eqivalent to or whether it's translation is already complete.

    Members:
    literal  -- The literal representing the theory atom.
    literals -- All literals equivalent to the theory atom.
    todo     -- Literals that have not yet been made equivalent to the
                representative literal.
    done     -- Whether translation of the theory atom is done.
    """
    def __init__(self):
        """
        Initialize the step information.
        """
        self.literal  = None
        self.literals = set()
        self.todo     = []
        self.done     = True

    def add_literal(self, backend):
        """
        Set the representative of the literal.

        Must only be called once. If there are already equivalent literals one
        of them is chosen as a representative.  Otherwise, a literal and a
        choice rule is added.

        Arguments:
        backend -- Backend to add the choice rule and literal to.
        """
        assert(self.literal is None)
        if len(self.literals) > 0:
            self.literal = min(self.literals)
            self.literals.remove(self.literal)
        else:
            self.literal = backend.add_atom()
            backend.add_rule([self.literal], [], True)
        return self.literal


class BodyFormula(Formula):
    """
    Base class of all temporal and Boolean formulas occurring in rule bodies.

    Members:
    __rep  -- unique string representation of the formula
    __data -- map from time points to StepData objects
    """
    def __init__(self, rep):
        """
        Initializes a formula with the given string representation.
        """
        self.__rep  = rep
        self.__data = {}

    @property
    def _rep(self):
        """
        Return the unique string representaiton of the formula.
        """
        return self.__rep

    def translate(self, ctx, step):
        """
        Translates a formula at a given step.

        Adds a new StepData object to the __data map (if it does not exist
        yet). Calls do_translate, which has to be implemented by base classes.
        And makes sure that the theory atom has a representative literal and
        all literals associated with the theory atom are made equivalent by
        rules.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        """
        data = self.__data.setdefault(step, StepData())
        self.do_translate(ctx, step, data)
        if len(data.todo) > 0:
            for atom in data.todo:
                make_equal(ctx.backend, atom, data.literal)
            del data.todo[:]
        return data.literal

    def add_atom(self, atom, step):
        """
        Adds the given atom to the equivalent literals of the theory atom at
        the given step.

        Arguments:
        atom -- ASP atom to add to the theory atom.
        step -- Step at which to add.
        """
        data = self.__data.setdefault(step, StepData())
        if atom not in data.literals:
            data.literals.add(atom)
            data.todo.append(atom)

# Boolean Formulas {{{1

class Atom(BodyFormula):
    """
    An atomic formula.

    Members:
    __name      -- Predicate name.
    __arguments -- Arguments of the atom (list of symbols).
    __positive  -- Classical negation sign.
    """
    def __init__(self, name, arguments=[], positive=True):
        """
        Initializes the atom.

        Arguments:
        name      -- Name of the atom.
        arguments -- Arguments of the atom.
        positive  -- Classical negation sign.
        """
        rep = "({}{}({}))".format("" if positive else "-", name, ",".join([str(a) for a in arguments]))
        if name.startswith("'"):
            raise RuntimeError("temporal formulas use < instead of leading primes: {}".format(rep))
        if name.endswith("'"):
            raise RuntimeError("temporal formulas use > instead of trailing primes: {}".format(rep))
        BodyFormula.__init__(self, rep)
        self.__name      = name
        self.__arguments = arguments
        self.__positive  = positive

    def do_translate(self, ctx, step, data):
        """
        Translates an atom.

        Requires that the step is within the horizon.

        This means looking the atom up in the symbolic atoms for the given step
        and setting the literal associated with the step accordingly. If the
        atom does not exist, it is set to false.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            sym = _clingo.Function(self.__name, self.__arguments + [_clingo.Number(step)], self.__positive)
            sym_atom = ctx.symbols[sym]
            data.literal = sym_atom.literal if sym_atom is not None else ctx.false_literal

class NumericLiteral(BodyFormula):
    """
    An formula over a numeric literal.

    Members:
    __literal -- The numeric literal.
    """
    def __init__(self, literal):
        """
        Initializes the literal.

        Arguments:
        literal -- The numeric literal.
        """
        BodyFormula.__init__(self, literal)
        self.__literal = literal

    def do_translate(self, ctx, step, data):
        """
        Translates the literal.

        Works the same as Atom.do_translate except that we already have a
        literal.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            data.literal = self.__literal


class BooleanConstant(BodyFormula):
    """
    Formula capturing a Boolean constant.

    Members:
    __value -- Truth value of the formula.
    """
    def __init__(self, value):
        """
        Initializes the formula with the given value.

        Members:
        __value -- Boolean value of the formula.
        """
        BodyFormula.__init__(self, "(&true)" if value else "(&false)")
        self.__value = value

    def do_translate(self, ctx, step, data):
        """
        Translates the formula.

        Requires that the step is within the horizon.

        Sets the literal of the data object to a true or false literal
        depending on the value of the formula.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            data.literal = -ctx.false_literal if self.__value else ctx.false_literal

class Negation(BodyFormula):
    """
    Formula capturing (default) negation.

    Members:
    __arg -- Formula to negate.
    """
    def __init__(self, arg):
        """
        Initializes the formula with the formula to negate.
        """
        BodyFormula.__init__(self, "(~{})".format(arg._rep))
        self.__arg = arg

    def do_translate(self, ctx, step, data):
        """
        Translates the formula.

        Requires that the step is within the horizon.

        Sets the literal to the negated literal of the subformula.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            data.literal = -self.__arg.translate(ctx, step)

class BooleanFormula(BodyFormula):
    """
    Formula capturing binary Boolean formulas.

    Members:
    __operator -- The Boolean connective.
    __lhs      -- The formula on the left-hand-side.
    __rhs      -- The formula on the left-hand-side.
    """

    def __init__(self, operator, lhs, rhs):
        """
        Initializes the formula.

        Arguments:
        operator -- Id of the connective.
        lhs      -- Formula on the left-hand-side.
        rhs      -- Formula on the right-hand-side.
        """
        rep = "({}{}{})".format(lhs._rep, operator, rhs._rep)
        BodyFormula.__init__(self, rep)
        self.__operator = operator
        self.__lhs      = lhs
        self.__rhs      = rhs

    def do_translate(self, ctx, step, data):
        """
        Translates the formula.

        Requires that the step is within the horizon.

        Sets a literal for the formula at the given step and adds clauses based
        on the type of the connective. Clauses are emulated with choices and
        integrity constraints.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            lhs = self.__lhs.translate(ctx, step)
            rhs = self.__rhs.translate(ctx, step)
            lit = data.add_literal(ctx.backend)
            if self.__operator != "<>":
                if self.__operator == "&":
                    lit, lhs, rhs = -lit, -lhs, -rhs
                elif self.__operator == "<-":
                    rhs = -rhs
                elif self.__operator == "->":
                    lhs = -lhs
                make_disjunction(ctx.backend, lit, lhs, rhs)
            elif self.__operator == "<>":
                ctx.backend.add_rule([], [-lit,  rhs,  lhs])
                ctx.backend.add_rule([], [-lit, -rhs, -lhs])
                ctx.backend.add_rule([], [ lit,  rhs, -lhs])
                ctx.backend.add_rule([], [ lit, -rhs,  lhs])

# Temporal Formulas {{{1

class Previous(BodyFormula):
    """
    Captures a formula referring to the previous state.

    Members:
    __arg  -- The argument of the previous operator.
    __weak -- Whether this is a weak previous operator.
    __n    -- How many steps to look back.
    """
    def __init__(self, arg, n, weak):
        """
        Initializes the formula.

        Arguments:
        arg  -- The argument of the previous operator.
        weak -- Whether this is a weak previous operator.
        n    -- How many steps to look back.
        """
        assert(n > 0)
        BodyFormula.__init__(self, "({}{}{})".format(n, "<:" if weak else "<", arg._rep))
        self.__arg  = arg
        self.__weak = weak
        self.__n = n

    def do_translate(self, ctx, step, data):
        """
        Translates an atom.

        Requires that the step is within the horizon.

        Translates the argument with respect to the previous step and sets the
        literal of the formula to the literal obtained thus. If the current
        step refers to the initial state, the literal is either set to the
        true or false literal depending on whether this is a weak previous
        operator or not.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            if step >= self.__n:
                data.literal = self.__arg.translate(ctx, step - self.__n)
            else:
                data.literal = ctx.false_literal
                if self.__weak and step < self.__n:
                    data.literal = -data.literal

class Initially(BodyFormula):
    """
    Captures a formula referring to the initial situation.

    Members:
    __arg  -- The argument of the previous operator.
    """
    def __init__(self, arg):
        """
        Initializes the formula.

        Arguments:
        arg  -- The argument of the initial operator.
        """
        BodyFormula.__init__(self, "(<<{})".format(arg._rep))
        self.__arg  = arg

    def do_translate(self, ctx, step, data):
        """
        Translates an atom.

        Requires that the step is within the horizon.

        Translates the argument with respect to the initial state and sets the
        literal accordingly.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            data.literal = self.__arg.translate(ctx, 0)

class Next(BodyFormula):
    """
    Captures a formula referring to the next state.

    Members:
    __arg  -- The argument of the next operator.
    __weak -- Whether this is a weak next operator.
    __n    -- How many steps to look ahead.
    """
    def __init__(self, arg, n, weak):
        """
        Initializes the formula.

        Arguments:
        arg  -- The argument of the next operator.
        weak -- Whether this is a weak next operator.
        n    -- How many steps to look ahead.
        """
        BodyFormula.__init__(self, "({}{}{})".format(n, ">:" if weak else ">", arg._rep))
        assert(n > 0)
        self.__arg  = arg
        self.__weak = weak
        self.__n    = n

    def do_translate(self, ctx, step, data):
        """
        Translates an atom.

        Requires that the step is within the horizon.

        Translates the argument with respect to the next step and sets the
        literal of the formula to the literal obtained thus. If the current
        step refers to the final state, a false external literal is created and
        the translation deferred until the next step.

        Note that the correctness of this translation requires that next
        operators are not used in rule heads, which is forbidden by the theory
        definition.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            if step + self.__n <= ctx.horizon:
                data.literal = self.__arg.translate(ctx, step + self.__n)
                data.done = True
            else:
                data.literal = ctx.backend.add_atom()
                true = getattr_(_clingo.TruthValue, "_True", "True_", "True")
                false = getattr_(_clingo.TruthValue, "_False", "False_", "False")
                ctx.backend.add_external(data.literal, true if self.__weak else false)
                ctx.add_todo(self, step)
                data.done = False
        elif not data.done:
            assert(step in range(0, ctx.horizon + 1))
            if step + self.__n <= ctx.horizon:
                arg = self.__arg.translate(ctx, step + self.__n)
                make_equal(ctx.backend, data.literal, arg)
                ctx.backend.add_external(data.literal, _clingo.TruthValue.Free)
                data.done = True
            else:
                ctx.add_todo(self, step)

class TelFormula(BodyFormula):
    """
    Captures a since, trigger, release, or until formula.

    This is an abstract class providing the basic translation for formulas
    referring to both the future and past.  Due to technical differences, the
    future and past versions are implemented in specialised classes.

    The left-hand-side of the operator can be None in which case either an
    eventually or an always operator is represented.

    Members:
    _op  -- The id of the operator.
    _lhs -- The left-hand-side of the temporal operator.
    _rhs -- The right-hand-side of the temporal operator.
    """

    def __init__(self, rep, op, lhs, rhs):
        """
        Initializes the formula.

        Arguments:
        rep -- Stringu representation of the formula.
        arg -- The id of the operator.
        lhs -- The left-hand-side of the operator.
        rhs -- The right-hand-side of the operator.
        """
        BodyFormula.__init__(self, rep)
        self._op  = op
        self._lhs = lhs
        self._rhs = rhs

    def _translate(self, ctx, step, data, pre):
        """
        Performs the translation of the temporal operator common to both future
        and past formulas.

        The translation works inductively. The literal pre is the literal
        obtained from the inductive step. Since since and trigger are dual, the
        same clauses are added but with the literals inverted in the trigger
        case.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        pre  -- The literal obtained from the inductive step.
        """
        lhs = None if self._lhs is None else self._lhs.translate(ctx, step)
        rhs = self._rhs.translate(ctx, step)
        lit = data.add_literal(ctx.backend)
        if self._op == "<*" or self._op == ">*":
            lit, rhs, pre = -lit, -rhs, -pre
            if lhs is not None:
                lhs = -lhs
        ctx.backend.add_rule([], [-lit, rhs])
        ctx.backend.add_rule([], [-rhs, -pre, lit])
        if lhs is not None:
            ctx.backend.add_rule([], [-lit,  lhs, pre])
            ctx.backend.add_rule([], [-rhs, -lhs, lit])
        else:
            ctx.backend.add_rule([], [-lit, pre])


class TelFormulaP(TelFormula):
    """
    Captures a since or trigger formulas.

    The left-hand-side of the operator can be None in which case either an
    eventually or an always operator is represented.
    """
    def __init__(self, op, lhs, rhs):
        """
        Initializes the formula.

        Arguments:
        arg -- The id of the operator.
        lhs -- The left-hand-side of the operator.
        rhs -- The right-hand-side of the operator.
        """
        rep = "({}{}{})".format("" if lhs is None else lhs._rep, op, rhs._rep)
        TelFormula.__init__(self, rep, op, lhs, rhs)

    def do_translate(self, ctx, step, data):
        """
        Translates the formula.

        Requires that the step is within the horizon.

        The formula is translated inductively using TelFormula._translate.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            if step == 0:
                data.literal = self._rhs.translate(ctx, step)
            else:
                pre = self.translate(ctx, step - 1)
                self._translate(ctx, step, data, pre)

class TelFormulaN(TelFormula):
    """
    Captures a release or until formulas.

    The left-hand-side of the operator can be None in which case either an
    eventually or an always operator is represented.

    Members:
    __future -- Next formula referring to the future to ease the translation.
    """
    def __init__(self, op, lhs, rhs):
        """
        Initializes the formula.

        Arguments:
        arg -- The id of the operator.
        lhs -- The left-hand-side of the operator.
        rhs -- The right-hand-side of the operator.
        """
        rep = "({}{}{})".format("" if lhs is None else lhs._rep, op, rhs._rep)
        TelFormula.__init__(self, rep, op, lhs, rhs)
        self.__future = None

    def set_future(self, future):
        """
        Set the future formula for the inductive translation.

        This is something like (> self).

        Arguments:
        future -- the formula
        """
        self.__future = future

    def do_translate(self, ctx, step, data):
        """
        Translates the formula.

        Requires that the step is within the horizon.

        The formula is translated inductively using TelFormula._translate.

        Arguments:
        ctx  -- Context object.
        step -- Step at which to translate.
        data -- Step data associated with the step.
        """
        if data.literal is None:
            assert(step in range(0, ctx.horizon + 1))
            fut = self.__future.translate(ctx, step)
            self._translate(ctx, step, data, fut)

# Dynamic Formulas {{{1

class DelFormula(BodyFormula):
    """
    Captures a diamond or box formula.

    Members:
    _op   -- The id of the operator.
    _path -- The left-hand-side of the temporal operator.
    _rhs  -- The right-hand-side of the temporal operator.
    """

    def __init__(self, rep, op, path, rhs):
        """
        Initializes the formula.

        Arguments:
        rep -- String representation of the formula.
        op -- The id of the operator.
        lhs -- The left-hand-side of the operator.
        rhs -- The right-hand-side of the operator.
        """
        self._op   = op
        self._path = path 
        self._rhs  = rhs
        BodyFormula.__init__(self, rep)

class DiamondFormula(DelFormula):
    def __init__(self, path, rhs):
        rep ="({}{}{}{})".format("<", path._rep, ">", rhs._rep)
        DelFormula.__init__(self, rep, "<>", path, rhs)

    def do_translate(self, ctx, step, data):
        if data.literal is None:
            attr = "translate_" + self._path.__class__.__name__
            data.add_literal(ctx.backend)
            getattr(self, attr)(ctx, step, data)

    def translate_ChoicePath(self,ctx, step, data):
        lhs = ctx.add_formula(DiamondFormula(self._path._lhs, self._rhs))
        rhs = ctx.add_formula(DiamondFormula(self._path._rhs, self._rhs))
        self.add_atom(ctx.add_formula(BooleanFormula("|", rhs, lhs)).translate(ctx, step), step)

    def translate_SequencePath(self, ctx, step, data):
        f = ctx.add_formula(DiamondFormula(self._path._rhs, self._rhs)) 
        self.add_atom(ctx.add_formula(DiamondFormula(self._path._lhs, f)).translate(ctx, step), step)
        
    def translate_CheckPath(self, ctx, step, data):
        self.add_atom(ctx.add_formula(BooleanFormula("&", self._path._arg, self._rhs)).translate(ctx, step), step)
        
    def translate_KleeneStarPath(self,ctx, step, data):
        final =  ctx.add_formula(Negation(ctx.add_formula(Next(ctx.add_formula(BooleanConstant(True)), 1, False))))
        a = ctx.add_formula(BooleanFormula("->", final, self._rhs))
        b = ctx.add_formula(BooleanFormula("|", self._rhs, ctx.add_formula(DiamondFormula(self._path._arg, self))))
        self.add_atom(ctx.add_formula(BooleanFormula("&", a, b)).translate(ctx, step), step)

    def translate_SkipPath(self,ctx, step, data):
        self.add_atom(ctx.add_formula(Next(self._rhs, 1, False)).translate(ctx, step), step)

class BoxFormula(DelFormula):
    def __init__(self, path, rhs):
        rep ="({}{}{}{})".format("[", path._rep,"]", rhs._rep)
        DelFormula.__init__(self, rep, "[]", path, rhs)

    def do_translate(self, ctx, step, data):
        if data.literal is None:
            attr = "translate_" + self._path.__class__.__name__
            data.add_literal(ctx.backend)
            getattr(self, attr)(ctx, step, data)

    def translate_ChoicePath(self,ctx, step, data):
        lhs = ctx.add_formula(BoxFormula(self._path._lhs, self._rhs))
        rhs = ctx.add_formula(BoxFormula(self._path._rhs, self._rhs))
        self.add_atom(ctx.add_formula(BooleanFormula("&", rhs, lhs)).translate(ctx, step), step)

    def translate_SequencePath(self,ctx, step, data):
        f = ctx.add_formula(BoxFormula(self._path._rhs, self._rhs)) 
        self.add_atom(ctx.add_formula(BoxFormula(self._path._lhs, f)).translate(ctx, step), step)
        
    def translate_CheckPath(self,ctx, step, data):
        self.add_atom(ctx.add_formula(BooleanFormula("->", self._path._arg, self._rhs)).translate(ctx, step), step)
        
    def translate_KleeneStarPath(self,ctx, step, data):
        final =  ctx.add_formula(Negation(ctx.add_formula(Next(ctx.add_formula(BooleanConstant(True)), 1, False))))
        a = ctx.add_formula(BooleanFormula("->", final, self._rhs))
        b = ctx.add_formula(BooleanFormula("&", self._rhs, ctx.add_formula(BoxFormula(self._path._arg,self))))
        self.add_atom(ctx.add_formula(BooleanFormula("&", a, b)).translate(ctx, step), step)

    def translate_SkipPath(self,ctx, step, data):
        self.add_atom(ctx.add_formula(Next(self._rhs, 1, True)).translate(ctx, step), step)

# Theory of Formulas {{{1

def create_atom(rep, add_formula, positive):
    """
    Returns the atom corresponding the given theory term.

    Throws an error if rep it is not a valid atom.

    Arguments:
    rep         -- Theory term to translate.
    add_formula -- Callback to add resulting formulas.
    positive    -- Boolean indicating the classical sign.
    """
    if rep.type == _clingo.TheoryTermType.Symbol:
        return add_formula(Atom(rep.name, [], positive))
    elif rep.type == _clingo.TheoryTermType.Function:
        if rep.name == "-" and len(rep.arguments) == 1:
            return create_atom(rep.arguments[0], add_formula, not positive)
        elif rep.name not in g_all_operators:
            return add_formula(Atom(rep.name, [create_symbol(arg) for arg in rep.arguments], positive))
    raise RuntimeError("invalid atom: {}".format(rep))

def create_path(rep, add_formula, check):
    """
    Returns the path formula corresponding the given path term.

    Throws an error if rep it is not a valid path formula.

    Arguments:
    rep         -- Term to translate.
    add_formula -- Callback to add resulting formuals.
    """
    if rep.type == _clingo.TheoryTermType.Symbol:
        if check:
            return create_atom(rep, add_formula, True)
        else:
            return add_formula(SequencePath(add_formula(CheckPath(create_atom(rep, add_formula, True))), add_formula(SkipPath())))
    elif rep.type == _clingo.TheoryTermType.Function:
        args = rep.arguments
        if rep.name in g_path_binary_operators:
            if check:
                raise RuntimeError("invalid dynamic formula: {}".format(rep))
            lhs = create_path(args[0], add_formula, False)
            rhs = create_path(args[1], add_formula, False)
            if rep.name == "+":
                return add_formula(ChoicePath(lhs,rhs))
            else:
                assert(rep.name == ";;")
                return add_formula(SequencePath(lhs, rhs))
        elif rep.name in g_path_unary_operators:
            if check:
                raise RuntimeError("invalid dynamic formula: {}".format(rep))
            if rep.name == "?":
                arg = create_path(args[0], add_formula, True)
                return add_formula(CheckPath(arg))
            else:
                assert(rep.name == "*")
                arg = create_path(args[0], add_formula, False)
                return add_formula(KleeneStarPath(arg))
        elif rep.name == "&":
            arg = rep.arguments[0]
            if arg.type == _clingo.TheoryTermType.Symbol:
                if not check and arg.name == "true":
                    return add_formula(SkipPath())
                if check and arg.name == "true" or arg.name == "false":
                    return add_formula(BooleanConstant(arg.name == "true"))
                else:
                    raise RuntimeError("unknown identifier: {}".format(rep))
            else:
                raise RuntimeError("invalid dynamic formula: {}".format(rep))
        #this case is probably impossible 
        else:
            return create_atom(rep, add_formula, True)
    else:
        raise RuntimeError("invalid dynamic formula: {}".format(rep))

def create_dynamic_formula(rep, add_formula):
    """
    Returns the dynamic formula corresponding the given theory term.

    Throws an error if rep it is not a valid formula.

    Arguments:
    rep         -- Theory term to translate.
    add_formula -- Callback to add resulting formuals.
    """
    if rep.type == _clingo.TheoryTermType.Symbol:
        return create_atom(rep, add_formula, True)
    elif rep.type == _clingo.TheoryTermType.Function:
        args = rep.arguments
        if rep.name in g_del_operators:
            lhs = create_path(args[0], add_formula, False)
            rhs = create_dynamic_formula(args[1], add_formula)
            if rep.name == ".>*":
                return add_formula(BoxFormula(lhs,rhs))
            else:
                assert(rep.name == ".>?")
                return add_formula(DiamondFormula(lhs, rhs))
        elif rep.name == "&":
            arg = rep.arguments[0]
            if arg.type == _clingo.TheoryTermType.Symbol:
                if arg.name == "true" or arg.name == "false":
                    return add_formula(BooleanConstant(arg.name == "true"))
                elif arg.name == "final":
                    return add_formula(BoxFormula(add_formula(SkipPath()), add_formula(BooleanConstant(False))))
                else:
                    raise RuntimeError("unknown identifier: {}".format(rep))
            else:
                raise RuntimeError("invalid dynamic formula: {}".format(rep))
        elif rep.name in g_all_operators:
            raise RuntimeError("invalid dynamic formula: {}".format(rep))
        else:
            return create_atom(rep, add_formula, True)
    else:
        raise RuntimeError("invalid dynamic formula: {}".format(rep))

def create_formula(rep, add_formula):
    """
    Returns the temporal formula corresponding the given theory term.

    Throws an error if rep it is not a valid formula.

    Arguments:
    rep         -- Theory term to translate.
    add_formula -- Callback to add resulting formuals.
    """
    if rep.type == _clingo.TheoryTermType.Symbol:
        return create_atom(rep, add_formula, True)
    elif rep.type == _clingo.TheoryTermType.Function:
        args = rep.arguments
        if rep.name in g_binary_operators and len(args) == 2:
            lhs = create_formula(args[0], add_formula)
            rhs = create_formula(args[1], add_formula)
            return add_formula(BooleanFormula(rep.name, lhs, rhs))
        elif rep.name in g_unary_operators and len(args) == 1:
            arg = create_formula(args[0], add_formula)
            return add_formula(Negation(arg))
        elif rep.name in g_tel_operators:
            rhs = create_formula(args[-1], add_formula)
            if rep.name == "<" or rep.name == "<:":
                lhs = 1 if len(args) == 1 else create_number(args[0])
                return rhs if lhs == 0 else add_formula(Previous(rhs, lhs, rep.name == "<:"))
            elif rep.name == ">" or rep.name == ">:":
                lhs = 1 if len(args) == 1 else create_number(args[0])
                return rhs if lhs == 0 else add_formula(Next(rhs, lhs, rep.name == ">:"))
            lhs = None if len(args) == 1 else create_formula(args[0], add_formula)
            if rep.name == "<;" or rep.name == "<:;":
                return add_formula(BooleanFormula("&", Previous(lhs, 1, rep.name == "<:;"), rhs))
            elif rep.name == "<*":
                return add_formula(TelFormulaP("<*", lhs, rhs))
            elif rep.name == "<?":
                return add_formula(TelFormulaP("<?", lhs, rhs))
            elif rep.name == "<<":
                return add_formula(Initially(rhs))
            elif rep.name == ";>" or rep.name == ";>:":
                return add_formula(BooleanFormula("&", lhs, Next(rhs, 1, rep.name == ";>:")))
            elif rep.name == ">*":
                formula = add_formula(TelFormulaN(">*", lhs, rhs))
                formula.set_future(add_formula(Next(formula, 1, True)))
                return formula
            elif rep.name == ">?":
                formula = add_formula(TelFormulaN(">?", lhs, rhs))
                formula.set_future(add_formula(Next(formula, 1, False)))
                return formula
            else:
                assert(rep.name == ">>")
                rhs = add_formula(BooleanFormula("|", add_formula(Negation(add_formula(Atom("__final", [], True)))), rhs))
                formula = add_formula(TelFormulaN(">*", None, rhs))
                formula.set_future(add_formula(Next(formula, 1, True)))
                return formula
        elif rep.name == "&":
            arg = rep.arguments[0]
            if arg.type == _clingo.TheoryTermType.Symbol:
                if arg.name == "initial" or arg.name == "final":
                    name = "__{}".format(arg.name)
                    return add_formula(Atom(name, [], True))
                elif arg.name == "true" or arg.name == "false":
                    return add_formula(BooleanConstant(arg.name == "true"))
                else:
                    raise RuntimeError("unknown identifier: {}".format(rep))
            else:
                raise RuntimeError("invalid temporal formula: {}".format(rep))
        else:
            return create_atom(rep, add_formula, True)
    else:
        raise RuntimeError("invalid temporal formula: {}".format(rep))

def translate_conjunction(formulas, add_formula):
    """
    Return a formula corresponding to the conjunction of the given
    formulas.

    Arguments:
    formulas    -- List of temporal formulas.
    add_formula -- Callback to add resulting formuals.
    """
    if len(formulas) == 0:
        return BooleanConstant(True)

    formulas.sort(key=lambda x: x._rep)
    formula = formulas[0]
    for x in formulas[1:]:
        formula = add_formula(BooleanFormula("&", formula, x))
    return formula


def translate_elements(elements, add_formula, dynamic):
    """
    Translate the given conjunction of elements and return a formula.

    This translation is corresponds to the handling of conditional literals
    in _clingo. The difference is that here only the classical case is
    supported.

    Arguments:
    elements    -- List of theory elements.
    add_formula -- Callback to add resulting formuals.
    """
    formulas = []

    for element in elements:
        if dynamic:
            formulas.append(create_dynamic_formula(element.terms[0], add_formula))
        else:
            formulas.append(create_formula(element.terms[0], add_formula))
        if len(element.condition) > 0:
            condition = translate_conjunction([add_formula(NumericLiteral(literal)) for literal in element.condition], add_formula)
            formulas[-1] = add_formula(BooleanFormula("->", condition, formulas[-1]))

    return translate_conjunction(formulas, add_formula)



"""
This module exports functions to translate (ground) theory atoms to rules via
clingo's backend.
"""

import clingo as _clingo
from . import formula as _frm
from . import body as _bd
from . import head as _hd

class Theory:
    """
    Class holding a set of formulas.

    Formulas can be added to this theory incrementally. Such formulas are then
    also incrementally translated into rules.

    Members:
    __formulas      -- A dictionary of formulas mapping string representations
                       to actual formulas.
    __todo_keys     -- Set of keys of formulas that still have to be translated
                       (makes sure that formulas in the todo list appear only
                       once).
    __todo          -- List of formulas to translate.
    __false_literal -- A literal that is false used during translation.
    """
    def __init__(self):
        """
        Initializes an empty theory.
        """
        self.__formulas = {}
        self.__todo_keys = set()
        self.__todo = []
        self.__false_literal = None

    def add_formula(self, formula):
        """
        Add the given formula to the theory.
        """
        formula = self.__formulas.setdefault(formula._rep, formula)
        return formula

    def add_todo(self, formula, step):
        """
        Add the given formula to the todo list.

        Arguments:
        formula -- The formula to add.
        step    -- The step at which to translate the formula.
        """
        key = (step, formula._rep)
        if key not in self.__todo_keys:
            self.__todo_keys.add(key)
            self.__todo.append((step, formula))

    def false_literal(self, backend):
        """
        Returns a false program literal.

        Arguments:
        backend -- The backend to add the literal to if there is no false
                   literal yet.
        """
        if self.__false_literal is None:
            self.__false_literal = backend.add_atom()
        return self.__false_literal

    def translate(self, horizon, prg):
        """
        Translates the next step for the given horizon.

        Also adds theory atoms from prg.

        Arguments:
        horizon -- The current horizon.
        prg     -- Control object (with theory atoms).
        """
        for atom in prg.theory_atoms:
            if atom.term.name == "del" and len(atom.term.arguments) == 1:
                step    = atom.term.arguments[0].number
                formula = _bd.translate_elements(atom.elements, self.add_formula, True)
                formula.add_atom(atom.literal, step)
                self.add_todo(formula, step)
            elif atom.term.name == "tel" and len(atom.term.arguments) == 1:
                step    = atom.term.arguments[0].number
                formula = _bd.translate_elements(atom.elements, self.add_formula, False)
                formula.add_atom(atom.literal, step)
                self.add_todo(formula, step)
            elif atom.term.name == "__tel_head" and len(atom.term.arguments) == 1:
                step    = atom.term.arguments[0].number
                formula = _hd.translate_formula(atom, self.add_formula)
                self.add_todo(formula, step)

        if len(self.__todo) > 0:
            todo, self.__todo, self.__todo_keys = self.__todo, [], set()
            with prg.backend() as b:
                ctx = _frm.Context(b, prg.symbolic_atoms, self.add_todo, self.add_formula, self.false_literal, horizon)
                for step, formula in todo:
                    formula.translate(ctx, step)



class Path(object):
    """
    Base class of all path

    Members:
    __rep  -- unique string representation of the path
    """
    def __init__(self, rep):
        """
        Initializes a formula with the given string representation.
        """
        self.__rep  = rep

    @property
    def _rep(self):
        """
        Return the unique string representaiton of the formula.
        """
        return self.__rep

class SkipPath(Path):
    def __init__(self):
        Path.__init__(self, "(&skip)")

class BinaryPath(Path):
    """
    Members:
    _lhs
    _rhs

    """
    def __init__(self, rep, lhs, rhs):
        self._lhs = lhs
        self._rhs = rhs
        Path.__init__(self, rep)

class ChoicePath(BinaryPath):
    
    def __init__(self, lhs, rhs):
        BinaryPath.__init__(self, "({}+{})".format(lhs._rep, rhs._rep), lhs, rhs)

class SequencePath(BinaryPath):
    
    def __init__(self, lhs, rhs):
        BinaryPath.__init__(self, "({};;{})".format(lhs._rep, rhs._rep), lhs, rhs)

class UnaryPath(Path):
    """
    Members:
    __arg
    """
    def __init__(self, rep, arg):
        self.__arg = arg 
        Path.__init__(self, rep)

    @property
    def _arg(self):
        """
        Return the unique string representaiton of the formula.
        """
        return self.__arg

class CheckPath(UnaryPath):
    
    def __init__(self, arg):
        UnaryPath.__init__(self, "({}?)".format(arg._rep), arg)

class KleeneStarPath(UnaryPath):

    def __init__(self, arg):
        UnaryPath.__init__(self, "({}*)".format(arg._rep), arg)
